/* Spectrum tools WSR interface (for Metageeks Chanalyzer 3.0)
 *
 * Basic dumper to dump the raw values from the USB device to a WSR file.
 * WSR are Wispy Recording Files as generated by the Chanalyser software.
 * 
 * By Frank Siegert, frank@wizards.de
 *
 * WILL ONLY WORK WITH WiSpy 2.4 AT THE MOMENT
 *
 * Based on spectool_raw by Mike Kershaw/Dragorn <dragorn@kismetwireless.net>
 *
 * This code is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * Extra thanks to Ryan Woodings @ Metageek for interface documentation
 */

#include <stdio.h>
#include <usb.h>
#include <stdlib.h>
#include <signal.h>
#include <getopt.h>
#include <string.h>
#include <sys/select.h>
#include <sys/param.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <errno.h>

#include "config.h"

#include "spectool_container.h"
#include "spectool_net_client.h"

wispy_phy *devs = NULL;
int ndev = 0;
int finish_write = 0;
int sampleCnt = 0;
char *wsr_filename = NULL;
FILE *wsr_out = NULL;

unsigned long amp_offset = 0;
unsigned long amp_res = 0;

void sighandle(int sig) {

	if (sig == SIGINT) {
		signal(SIGINT, sighandle);
		finish_write = 1;
		return;
	}

	printf("Dying %d from signal %d\n", getpid(), sig);

	exit(1);
}

void Usage(void) {
	printf("wispy_wsr [ options ] output.wsr\n"
		   " -n / --net  tcp://host:port  Connect to network server instead of\n"
		   " -b / --broadcast             Listen for (and connect to) broadcast servers\n"
		   "                              local USB devices\n");
	return;
}

void outputLong(unsigned long n) {
	unsigned long nv;
#ifndef WORDS_BIGENDIAN
	nv = htonl(n);	
#else
	nv = n;
#endif	
	fwrite(&nv,1,sizeof(unsigned long),wsr_out);
}

void outputWord(unsigned short n) {
	unsigned short nv;
#ifndef WORDS_BIGENDIAN
        nv = htons(n);
#else
        nv = n;
#endif
	fwrite(&nv,1,sizeof(unsigned short),wsr_out);
}

void outputByte(unsigned char n) {
	fputc(n, wsr_out);
}


void writeHeader(void) {
	/* this header is WiSpy 2.4 specific */

	outputLong(0x00000003);	   /* wsr version */
	outputWord(0x0002);        /* device ID 2 (WiSpy 2.4) */
	outputLong(0x00249EC2);	   /* FreqOffsetKHz */
	outputLong(0x00050106);	   /* FreqResHz */
	outputWord(0x0100);	   /* NumSamples */
	outputLong(0xFFFDF490);	   /* AmpOffset */
	outputLong(0x000001F4);	   /* AmpRes */
	outputWord(0x0109);	   /* RSSIMaxSteps */
	outputLong(0x00000000);	   /* First Bookmark File Pos (absolute) */
	outputLong(0x00000000);	
}

void writeFooterAndCloseWSR(void) {
	int i;
	unsigned long pos;

	/* static footer, so Chanalyer is happy */

	unsigned char footer[]={ 0x01, 0x00, 0x39, 0x47, 0x72, 0x9B, 0xED, 0x00, 0x02, \
	0xBB, 0x38, 0x73, 0x70, 0x65, 0x63, 0x74, 0x6F, 0x6F, \
	0x6C, 0x5F, 0x77, 0x73, 0x72, 0x20, 0x72, 0x65, 0x63, \
	0x6F, 0x72, 0x64, 0x69, 0x6E, 0x67, 0x20, 0x77, 0x69, \
	0x74, 0x68, 0x20, 0x57, 0x69, 0x53, 0x70, 0x79, 0x20, \
	0x32, 0x2E, 0x34, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, \
	0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x01, 0x00, 0x1C, \
	0x47, 0x72, 0x9C, 0x3A, 0x00, 0x02, 0xCA, 0xD8, 0x6E, \
	0x6F, 0x20, 0x69, 0x6E, 0x74, 0x65, 0x72, 0x6E, 0x65, \
	0x74, 0x20, 0x61, 0x63, 0x74, 0x69, 0x76, 0x69, 0x74, \
	0x79, 0x01, 0x00, 0x24, 0x47, 0x72, 0x9C, 0x5C, 0x00, \
	0x0B, 0xDB, 0x28, 0x54, 0x72, 0x65, 0x6E, 0x74, 0x20, \
	0x6F, 0x70, 0x65, 0x6E, 0x73, 0x20, 0x69, 0x6E, 0x74, \
	0x65, 0x72, 0x6E, 0x65, 0x74, 0x20, 0x62, 0x72, 0x6F, \
	0x77, 0x73, 0x65, 0x72, 0x01, 0x00, 0x32, 0x47, 0x72, \
	0x9D, 0x17, 0x00, 0x0E, 0xE8, 0x68, 0x54, 0x72, 0x65, \
	0x6E, 0x74, 0x20, 0x42, 0x65, 0x67, 0x69, 0x6E, 0x73, \
	0x20, 0x4D, 0x69, 0x63, 0x72, 0x6F, 0x77, 0x61, 0x76, \
	0x65, 0x20, 0x28, 0x53, 0x65, 0x65, 0x20, 0x53, 0x70, \
	0x65, 0x63, 0x74, 0x72, 0x61, 0x6C, 0x20, 0x56, 0x69, \
	0x65, 0x77, 0x29, 0xFF };

	fflush(wsr_out);
	pos = ftell(wsr_out);
	i = 0;
	while (footer[i] != 0xFF) {
		fputc(footer[i++], wsr_out);
	}
	fclose(wsr_out);
	wsr_out = fopen(wsr_filename, "r+");
	if (!wsr_out) {
		printf("Unable to update wsr file\n");
		exit(-1);
	}
	fseek(wsr_out,16,SEEK_SET);
	outputLong(amp_offset);
	outputLong(amp_res);

	fseek(wsr_out,30,SEEK_SET);  /* update header */
	outputLong(pos);

	fclose(wsr_out);
}			

void writeTimestamp(void) {
	struct timeval tv;
	gettimeofday(&tv,NULL);
	outputLong(tv.tv_sec);
	outputLong(tv.tv_usec);	
}

int main(int argc, char *argv[]) {
	wispy_device_list list;
	int x = 0, r = 0;
	wispy_sample_sweep *sb;
	spectool_server sr;
	char errstr[WISPY_ERROR_MAX];
	int ret;
	wispy_phy *pi;

	static struct option long_options[] = {
		{ "net", required_argument, 0, 'n' },
		{ "broadcast", no_argument, 0, 'b' },
		{ "help", no_argument, 0, 'h' },
		{ 0, 0, 0, 0 }
	};
	int option_index;

	char *neturl = NULL;

	char bcasturl[SPECTOOL_NETCLI_URL_MAX];
	int bcastlisten = 0;
	int bcastsock;

	while (1) {
		int o = getopt_long(argc, argv, "n:bh",
							long_options, &option_index);

		if (o < 0)
			break;

		if (o == 'h') {
			Usage();
			return;
		} else if (o == 'b') {
			bcastlisten = 1;
		} else if (o == 'n') {
			neturl = strdup(optarg);
			printf("debug - wispy_raw neturl %s\n", neturl);
			continue;
		}
	}

	if (optind < argc) {
		wsr_out = fopen(argv[optind],"wb");
		if (wsr_out == NULL) {
			printf("Error: Cannot write '%s'\n", argv[optind]);
			exit(-1);
		}
		wsr_filename = argv[optind];
		optind++;
	}

	if (wsr_out == NULL) {
		printf("No output specified\n");
		Usage();
		exit(-1);
	}

	signal(SIGINT, sighandle);

	if (bcastlisten) {
		printf("Initializing broadcast listen...\n");

		if ((bcastsock = spectool_netcli_initbroadcast(WISPY_NET_DEFAULT_PORT,
													   errstr)) < 0) {
			printf("Error initializing bcast socket: %s\n", errstr);
			exit(1);
		}

		printf("Waiting for a broadcast server ID...\n");
	} else if (neturl != NULL) {
		printf("Initializing network connection...\n");

		if (spectool_netcli_init(&sr, neturl, errstr) < 0) {
			printf("Error initializing network connection: %s\n", errstr);
			exit(1);
		}

		if (spectool_netcli_connect(&sr, errstr) < 0) {
			printf("Error opening network connection: %s\n", errstr);
			exit(1);
		}

		printf("Connected to server, waiting for device list...\n");
	} else if (neturl == NULL) {
		ndev = wispy_device_scan(&list);
		if (ndev <= 0) {
			printf("No wispy devices found, bailing\n");
			exit(1);
		}

		printf("Found %d wispy devices...\n", ndev);

		for (x = 0; x < ndev; x++) {
			printf("Initializing WiSPY device %s id %u\n", 
				   list.list[x].name, list.list[x].device_id);

			pi = (wispy_phy *) malloc(WISPY_PHY_SIZE);
			pi->next = devs;
			devs = pi;

			if (wispy_device_init(pi, &(list.list[x])) < 0) {
				printf("Error initializing WiSPY device %s id %u\n",
					   list.list[x].name, list.list[x].device_id);
				printf("%s\n", wispy_get_error(pi));
				exit(1);
			}

			if (wispy_phy_open(pi) < 0) {
				printf("Error opening WiSPY device %s id %u\n",
					   list.list[x].name, list.list[x].device_id);
				printf("%s\n", wispy_get_error(pi));
				exit(1);
			}

			wispy_phy_setcalibration(pi, 1);
		}

		wispy_device_scan_free(&list); 
	}

	writeHeader();

	/* Naive poll that doesn't use select() to find pending data */
	while (finish_write == 0) {
		fd_set rfds;
		fd_set wfds;
		int maxfd = 0;
		struct timeval tm;

		FD_ZERO(&rfds);
		FD_ZERO(&wfds);

		pi = devs;
		while (pi != NULL) {
			if (wispy_phy_getpollfd(pi) >= 0) {
				FD_SET(wispy_phy_getpollfd(pi), &rfds);

				if (wispy_phy_getpollfd(pi) > maxfd)
					maxfd = wispy_phy_getpollfd(pi);
			}

			pi = pi->next;
		}

		if (neturl != NULL) {
			if (spectool_netcli_getpollfd(&sr) >= 0) {
				FD_SET(spectool_netcli_getpollfd(&sr), &rfds);

				if (spectool_netcli_getpollfd(&sr) > maxfd)
					maxfd = spectool_netcli_getpollfd(&sr);
			}
			if (spectool_netcli_getwritepend(&sr) > 0) {
				FD_SET(spectool_netcli_getwritefd(&sr), &wfds);

				if (spectool_netcli_getwritefd(&sr) > maxfd)
					maxfd = spectool_netcli_getwritefd(&sr);
			}
		}

		if (bcastlisten) {
			FD_SET(bcastsock, &rfds);
			if (bcastsock > maxfd)
				maxfd = bcastsock;
		}

		tm.tv_sec = 0;
		tm.tv_usec = 5000;  /* timestep 500ms */

		if (select(maxfd + 1, &rfds, &wfds, NULL, &tm) < 0) {
			// printf("wispy_raw select() error: %s\n", strerror(errno));
			// exit(1);
			break;
		}

		if (bcastlisten && FD_ISSET(bcastsock, &rfds)) {
			if (spectool_netcli_pollbroadcast(bcastsock, bcasturl, errstr) == 1) {
				printf("Saw broadcast for server %s\n", bcasturl);

				if (neturl == NULL) {
					neturl = strdup(bcasturl);

					if (spectool_netcli_init(&sr, neturl, errstr) < 0) {
						printf("Error initializing network connection: %s\n", errstr);
						exit(1);
					}

					if (spectool_netcli_connect(&sr, errstr) < 0) {
						printf("Error opening network connection: %s\n", errstr);
						exit(1);
					}
				}
			}
		}

		if (spectool_netcli_getwritefd(&sr) >= 0 &&
			FD_ISSET(spectool_netcli_getwritefd(&sr), &wfds)) {
			if (spectool_netcli_writepoll(&sr, errstr) < 0) {
				printf("Error write-polling network server %s\n", errstr);
				exit(1);
			}
		}

		ret = SPECTOOL_NETCLI_POLL_ADDITIONAL;
		while (spectool_netcli_getpollfd(&sr) >= 0 &&
			   FD_ISSET(spectool_netcli_getpollfd(&sr), &rfds) &&
			   (ret & SPECTOOL_NETCLI_POLL_ADDITIONAL)) {

			if ((ret = spectool_netcli_poll(&sr, errstr)) < 0) {
				printf("Error polling network server %s\n", errstr);
				exit(1);
			}

			if ((ret & SPECTOOL_NETCLI_POLL_NEWDEVS)) {
				spectool_net_dev *ndi = sr.devlist;
				while (ndi != NULL) {
					printf("Enabling network device: %s (%u)\n", ndi->device_name,
						   ndi->device_id);
					pi = spectool_netcli_enabledev(&sr, ndi->device_id, errstr);

					pi->next = devs;
					devs = pi;

					ndi = ndi->next;
				}

			}
		}

		pi = devs;
		while (pi != NULL) {
			wispy_phy *di = pi;
			pi = pi->next;

			if (wispy_phy_getpollfd(di) < 0) {
				if (wispy_get_state(di) == WISPY_STATE_ERROR) {
					printf("Error polling wispy device %s\n",
						   wispy_phy_getname(di));
					printf("%s\n", wispy_get_error(di));
					exit(1);
				}
			}

			if (FD_ISSET(wispy_phy_getpollfd(di), &rfds) == 0) {
				continue;
			}

			do {
				r = wispy_phy_poll(di);

				if ((r & WISPY_POLL_CONFIGURED)) {
					printf("Configured device %u (%s): %d ranges\n", 
						   wispy_phy_getdevid(di), 
						   wispy_phy_getname(di),
						   di->device_spec->num_sweep_ranges);
					for (r = 0; r < di->device_spec->num_sweep_ranges; r++) {
						wispy_sample_sweep *ran = 
							&(di->device_spec->supported_ranges[r]);

						/* Don't do this, kids. */
						printf("    %d: %d%s-%d%s @ %d%s, %d samples\n", r, 
							   ran->start_khz > 1000 ? 
							   ran->start_khz / 1000 : ran->start_khz,
							   ran->start_khz > 1000 ? "MHz" : "KHz",
							   ran->end_khz > 1000 ? ran->end_khz / 1000 : ran->end_khz,
							   ran->end_khz > 1000 ? "MHz" : "KHz",
							   (ran->res_hz / 1000) > 1000 ? (ran->res_hz / 1000) / 1000 : ran->res_hz / 1000,
							   (ran->res_hz / 1000) > 1000 ? "MHz" : "KHz",
							   di->device_spec->supported_ranges[r].num_samples);
						printf("    %d: Offset %d res %u\n",
							   r, ran->amp_offset_mdbm, ran->amp_res_mdbm);
					}

					continue;
				} else if ((r & WISPY_POLL_ERROR)) {
					printf("Error polling wispy device %s\n",
						   wispy_phy_getname(di));
					printf("%s\n", wispy_get_error(di));
					exit(1);
				} else if ((r & WISPY_POLL_SWEEPCOMPLETE)) {
					sb = wispy_phy_getsweep(di);
					if (sb == NULL)
						continue;
/*
					printf("%s sweep off %d res %u: ", wispy_phy_getname(di),
						   sb->amp_offset_mdbm, sb->amp_res_mdbm);
					for (r = 0; r < sb->num_samples; r++) {
						printf("%d ", 
							WISPY_RSSI_CONVERT(sb->amp_offset_mdbm, sb->amp_res_mdbm,
											   sb->sample_data[r]));
					}
					printf("\n");
*/
					writeTimestamp(); sampleCnt++;
					
					amp_offset = sb->amp_offset_mdbm;
					amp_res =  sb->amp_res_mdbm;

					for (r = 0; r < sb->num_samples; r++) {
						outputByte(sb->sample_data[r]);
						// outputByte(WISPY_RSSI_CONVERT(sb->amp_offset_mdbm, sb->amp_res_mdbm, sb->sample_data[r]));
					}
				}
			} while ((r & WISPY_POLL_ADDITIONAL));

		}
	}

	writeFooterAndCloseWSR();

	wsr_out = NULL;

	printf("%d samples written\n", sampleCnt);

	return 0;
}	

